import os
import sys
import stat
import json
import shutil
import getpass
import logging
from time import sleep
from pathlib import Path
from threading import Lock
from itertools import chain
from contextlib import suppress
from secrets import token_bytes
from ansible_runner.interface import run
from subprocess import CalledProcessError

from ..misc import can_sudo_without_password, os_platform, rm_at_exit, get_python_constraints

log = logging.getLogger("bbot.core.helpers.depsinstaller")


class DepsInstaller:
    CORE_DEPS = {
        # core BBOT dependencies in the format of binary: package_name
        # each one will only be installed if the binary is not found
        "unzip": "unzip",
        "zipinfo": "unzip",
        "curl": "curl",
        "git": "git",
        "make": "make",
        "gcc": "gcc",
        "bash": "bash",
        "which": "which",
        "tar": "tar",
        # debian why are you like this
        "7z": [
            {
                "name": "Install 7zip (Debian)",
                "package": {"name": ["p7zip-full"], "state": "present"},
                "become": True,
                "when": "ansible_facts['os_family'] == 'Debian'",
            },
            {
                "name": "Install 7zip (Non-Debian)",
                "package": {"name": ["p7zip"], "state": "present"},
                "become": True,
                "when": "ansible_facts['os_family'] != 'Debian'",
            },
            {
                "name": "Install p7zip-plugins (Fedora)",
                "package": {"name": ["p7zip-plugins"], "state": "present"},
                "become": True,
                "when": "ansible_facts['distribution'] == 'Fedora'",
            },
        ],
    }

    def __init__(self, parent_helper):
        self.parent_helper = parent_helper
        self.preset = self.parent_helper.preset
        self.core = self.preset.core

        self.os_platform = os_platform()

        # respect BBOT's http timeout
        self.web_config = self.parent_helper.config.get("web", {})
        http_timeout = self.web_config.get("http_timeout", 30)
        os.environ["ANSIBLE_TIMEOUT"] = str(http_timeout)

        # cache encrypted sudo pass
        self.askpass_filename = "sudo_askpass.py"
        self._sudo_password = None
        self._sudo_cache_setup = False
        self._setup_sudo_cache()
        self._installed_sudo_askpass = False

        self.data_dir = self.parent_helper.cache_dir / "depsinstaller"
        self.parent_helper.mkdir(self.data_dir)
        self.setup_status_cache = self.data_dir / "setup_status.json"
        self.command_status = self.data_dir / "command_status"
        self.parent_helper.mkdir(self.command_status)
        self.setup_status = self.read_setup_status()

        # make sure we're using a minimal git config
        self.minimal_git_config = self.data_dir / "minimal_git.config"
        self.minimal_git_config.touch()
        os.environ["GIT_CONFIG_GLOBAL"] = str(self.minimal_git_config)

        self.deps_config = self.parent_helper.config.get("deps", {})
        self.deps_behavior = self.deps_config.get("behavior", "abort_on_failure").lower()
        self.ansible_debug = self.core.logger.log_level <= logging.DEBUG
        self.venv = ""
        if sys.prefix != sys.base_prefix:
            self.venv = sys.prefix

        self.ensure_root_lock = Lock()

    async def install(self, *modules):
        await self.install_core_deps()
        succeeded = []
        failed = []
        try:
            notified = False
            for m in modules:
                # assume success if we're ignoring dependencies
                if self.deps_behavior == "disable":
                    succeeded.append(m)
                    continue
                # abort if module name is unknown
                if m not in self.all_modules_preloaded:
                    log.verbose(f'Module "{m}" not found')
                    failed.append(m)
                    continue
                preloaded = self.all_modules_preloaded[m]
                log.debug(f"Installing {m} - Preloaded Deps {preloaded['deps']}")
                # make a hash of the dependencies and check if it's already been handled
                # take into consideration whether the venv or bbot home directory changes
                module_hash = self.parent_helper.sha1(
                    json.dumps(preloaded["deps"], sort_keys=True)
                    + self.venv
                    + str(self.parent_helper.bbot_home)
                    + os.uname()[1]
                ).hexdigest()
                success = self.setup_status.get(module_hash, None)
                dependencies = list(chain(*preloaded["deps"].values()))
                if len(dependencies) <= 0:
                    log.debug(f'No dependency work to do for module "{m}"')
                    succeeded.append(m)
                    continue
                else:
                    if (
                        success is None
                        or (success is False and self.deps_behavior == "retry_failed")
                        or self.deps_behavior == "force_install"
                    ):
                        if not notified:
                            log.hugeinfo("Installing module dependencies. Please be patient, this may take a while.")
                            notified = True
                        log.verbose(f'Installing dependencies for module "{m}"')
                        # get sudo access if we need it
                        if preloaded.get("sudo", False) is True:
                            self.ensure_root(f'Module "{m}" needs root privileges to install its dependencies.')
                        success = await self.install_module(m)
                        self.setup_status[module_hash] = success
                        if success or self.deps_behavior == "ignore_failed":
                            log.debug(f'Setup succeeded for module "{m}"')
                            succeeded.append(m)
                        else:
                            log.warning(f'Setup failed for module "{m}"')
                            failed.append(m)
                    else:
                        if success or self.deps_behavior == "ignore_failed":
                            log.debug(
                                f'Skipping dependency install for module "{m}" because it\'s already done (--force-deps to re-run)'
                            )
                            succeeded.append(m)
                        else:
                            log.warning(
                                f'Skipping dependency install for module "{m}" because it failed previously (--retry-deps to retry or --ignore-failed-deps to ignore)'
                            )
                            failed.append(m)

        finally:
            self.write_setup_status()

        succeeded.sort()
        failed.sort()
        return succeeded, failed

    async def install_module(self, module):
        success = True
        preloaded = self.all_modules_preloaded[module]

        # ansible tasks
        ansible_tasks = preloaded["deps"]["ansible"]
        if ansible_tasks:
            success &= self.tasks(module, ansible_tasks)

        # apt
        deps_apt = preloaded["deps"]["apt"]
        if deps_apt:
            self.apt_install(deps_apt)

        # shell
        deps_shell = preloaded["deps"]["shell"]
        if deps_shell:
            success &= self.shell(module, deps_shell)

        # pip
        deps_pip = preloaded["deps"]["pip"]
        deps_pip_constraints = preloaded["deps"]["pip_constraints"]
        if deps_pip:
            success &= await self.pip_install(deps_pip, constraints=deps_pip_constraints)

        # shared/common
        deps_common = preloaded["deps"]["common"]
        if deps_common:
            for dep_common in deps_common:
                if self.setup_status.get(dep_common, False) is True:
                    log.debug(
                        f'Skipping installation of dependency "{dep_common}" for module "{module}" since it is already installed'
                    )
                    continue
                ansible_tasks = self.preset.module_loader._shared_deps[dep_common]
                result = self.tasks(module, ansible_tasks)
                self.setup_status[dep_common] = result
                success &= result

        return success

    async def pip_install(self, packages, constraints=None):
        packages_str = ",".join(packages)
        log.info(f"Installing the following pip packages: {packages_str}")

        command = [sys.executable, "-m", "pip", "install", "--upgrade"] + packages

        # if no custom constraints are provided, use the constraints of the currently installed version of bbot
        if constraints is not None:
            constraints = get_python_constraints()

        constraints_tempfile = self.parent_helper.tempfile(constraints, pipe=False)
        command.append("--constraint")
        command.append(constraints_tempfile)

        process = None
        try:
            process = await self.parent_helper.run(command, check=True)
            message = f'Successfully installed pip packages "{packages_str}"'
            output = process.stdout
            if output is not None:
                message = output.splitlines()[-1]
            log.info(message)
            return True
        except CalledProcessError as err:
            log.warning(f"Failed to install pip packages {packages_str} (return code {err.returncode}): {err.stderr}")
        return False

    def apt_install(self, packages):
        """
        Install packages with the OS's default package manager (apt, pacman, dnf, etc.)
        """
        args, kwargs = self._make_apt_ansible_args(packages)
        success, err = self.ansible_run(module="package", args=args, **kwargs)
        if success:
            log.info(f'Successfully installed OS packages "{",".join(sorted(packages))}"')
        else:
            log.warning(
                f"Failed to install OS packages ({err}). Recommend installing the following packages manually:"
            )
            for p in packages:
                log.warning(f" - {p}")
        return success

    def _make_apt_ansible_args(self, packages):
        packages_str = ",".join(sorted(packages))
        log.info(f"Installing the following OS packages: {packages_str}")
        args = {"name": packages_str, "state": "present"}  # , "update_cache": True, "cache_valid_time": 86400}
        kwargs = {}
        # don't sudo brew
        if self.os_platform != "darwin":
            kwargs = {
                "ansible_args": {
                    "ansible_become": True,
                    "ansible_become_method": "sudo",
                }
            }
        return args, kwargs

    def shell(self, module, commands):
        tasks = []
        for i, command in enumerate(commands):
            command_hash = self.parent_helper.sha1(f"{module}_{i}_{command}").hexdigest()
            command_status_file = self.command_status / command_hash
            if type(command) == str:
                command = {"cmd": command}
            command["cmd"] += f" && touch {command_status_file}"
            tasks.append(
                {
                    "name": f"{module}.deps_shell step {i + 1}",
                    "ansible.builtin.shell": command,
                    "args": {"executable": "/bin/bash", "creates": str(command_status_file)},
                }
            )
        success, err = self.ansible_run(tasks=tasks)
        if success:
            log.info(f"Successfully ran {len(commands):,} shell commands")
        else:
            log.warning("Failed to run shell dependencies")
        return success

    def tasks(self, module, tasks):
        log.info(f"Running {len(tasks):,} Ansible tasks for {module}")
        success, err = self.ansible_run(tasks=tasks)
        if success:
            log.info(f"Successfully ran {len(tasks):,} Ansible tasks for {module}")
        else:
            log.warning(f"Failed to run Ansible tasks for {module}")
        return success

    def ansible_run(self, tasks=None, module=None, args=None, ansible_args=None):
        _ansible_args = {"ansible_connection": "local", "ansible_python_interpreter": sys.executable}
        if ansible_args is not None:
            _ansible_args.update(ansible_args)
        module_args = None
        if args:
            module_args = " ".join([f'{k}="{v}"' for k, v in args.items()])
        log.debug(f"ansible_run(module={module}, args={args}, ansible_args={ansible_args})")
        playbook = None
        if tasks:
            for task in tasks:
                if "package" in task:
                    # special case for macos
                    if self.os_platform == "darwin":
                        # don't sudo brew
                        task["become"] = False
                        # brew doesn't support update_cache
                        task["package"].pop("update_cache", "")
            playbook = {"hosts": "all", "tasks": tasks}
            log.debug(json.dumps(playbook, indent=2))
        if self._sudo_password is not None:
            _ansible_args["ansible_become_password"] = self._sudo_password
        playbook_hash = self.parent_helper.sha1(str(playbook)).hexdigest()
        data_dir = self.data_dir / (module if module else f"playbook_{playbook_hash}")
        shutil.rmtree(data_dir, ignore_errors=True)
        self.parent_helper.mkdir(data_dir)

        res = run(
            playbook=playbook,
            private_data_dir=str(data_dir),
            host_pattern="localhost",
            inventory={
                "all": {"hosts": {"localhost": _ansible_args}},
            },
            module=module,
            module_args=module_args,
            quiet=True,
            verbosity=0,
            cancel_callback=lambda: None,
        )

        log.debug(f"Ansible status: {res.status}")
        log.debug(f"Ansible return code: {res.rc}")
        success = res.status == "successful"
        err = ""
        for e in res.events:
            if self.ansible_debug and not success:
                log.debug(json.dumps(e, indent=2))
            if e["event"] == "runner_on_failed":
                err = e["event_data"]["res"]["msg"]
                break
        return success, err

    def read_setup_status(self):
        setup_status = {}
        if self.setup_status_cache.is_file():
            with open(self.setup_status_cache) as f:
                with suppress(Exception):
                    setup_status = json.load(f)
        return setup_status

    def write_setup_status(self):
        with open(self.setup_status_cache, "w") as f:
            json.dump(self.setup_status, f)

    def ensure_root(self, message=""):
        self._install_sudo_askpass()
        # skip if we've already done this
        if self._sudo_password is not None:
            return
        with self.ensure_root_lock:
            # first check if the environment variable is set
            _sudo_password = os.environ.get("BBOT_SUDO_PASS", None)
            if _sudo_password is not None or os.geteuid() == 0 or can_sudo_without_password():
                # if we're already root or we can sudo without a password, there's no need to prompt
                return

            if message:
                log.warning(message)
            while not self._sudo_password:
                # sleep for a split second to flush previous log messages
                sleep(0.1)
                _sudo_password = getpass.getpass(prompt="[USER] Please enter sudo password: ")
                if self.parent_helper.verify_sudo_password(_sudo_password):
                    log.success("Authentication successful")
                    self._sudo_password = _sudo_password
                else:
                    log.warning("Incorrect password")

    async def install_core_deps(self):
        to_install = set()
        to_install_friendly = set()
        playbook = []
        self._install_sudo_askpass()
        # ensure tldextract data is cached
        self.parent_helper.tldextract("evilcorp.co.uk")
        # install any missing commands
        for command, package_name_or_playbook in self.CORE_DEPS.items():
            if not self.parent_helper.which(command):
                to_install_friendly.add(command)
                if isinstance(package_name_or_playbook, str):
                    to_install.add(package_name_or_playbook)
                else:
                    playbook.extend(package_name_or_playbook)
        # install ansible community.general collection
        if not self.setup_status.get("ansible:community.general", False):
            log.info("Installing Ansible Community General Collection")
            try:
                command = ["ansible-galaxy", "collection", "install", "community.general"]
                await self.parent_helper.run(command, check=True)
                self.setup_status["ansible:community.general"] = True
                log.info("Successfully installed Ansible Community General Collection")
            except CalledProcessError as err:
                log.warning(
                    f"Failed to install Ansible Community.General Collection (return code {err.returncode}): {err.stderr}"
                )
        # construct ansible playbook
        if to_install:
            playbook.append(
                {
                    "name": "Install Core BBOT Dependencies",
                    "package": {"name": list(to_install), "state": "present"},
                    "become": True,
                }
            )
        # run playbook
        if playbook:
            log.info(f"Installing core BBOT dependencies: {','.join(sorted(to_install_friendly))}")
            self.ensure_root()
            self.ansible_run(tasks=playbook)

    def _setup_sudo_cache(self):
        if not self._sudo_cache_setup:
            self._sudo_cache_setup = True
            # write temporary encryption key, to be deleted upon scan completion
            self._sudo_temp_keyfile = self.parent_helper.temp_filename()
            # remove it at exit
            rm_at_exit(self._sudo_temp_keyfile)
            # generate random 32-byte key
            random_key = token_bytes(32)
            # write key to file and set secure permissions
            self._sudo_temp_keyfile.write_bytes(random_key)
            self._sudo_temp_keyfile.chmod(0o600)
            # export path to environment variable, for use in askpass script
            os.environ["BBOT_SUDO_KEYFILE"] = str(self._sudo_temp_keyfile.resolve())

    @property
    def encrypted_sudo_pw(self):
        if self._sudo_password is None:
            return ""
        return self._encrypt_sudo_pw(self._sudo_password)

    def _encrypt_sudo_pw(self, pw):
        from Crypto.Cipher import AES
        from Crypto.Util.Padding import pad

        key = self._sudo_temp_keyfile.read_bytes()
        cipher = AES.new(key, AES.MODE_CBC)
        ct_bytes = cipher.encrypt(pad(pw.encode(), AES.block_size))
        iv = cipher.iv.hex()
        ct = ct_bytes.hex()
        return f"{iv}:{ct}"

    def _install_sudo_askpass(self):
        if not self._installed_sudo_askpass:
            self._installed_sudo_askpass = True
            # install custom askpass script
            askpass_src = Path(__file__).resolve().parent / self.askpass_filename
            askpass_dst = self.parent_helper.tools_dir / self.askpass_filename
            shutil.copy(askpass_src, askpass_dst)
            askpass_dst.chmod(askpass_dst.stat().st_mode | stat.S_IEXEC)

    @property
    def all_modules_preloaded(self):
        return self.preset.module_loader.preloaded()



from bbot.modules.base import BaseModule


class IP2Location(BaseModule):
    """
    IP2Location.io Geolocation API.
    """

    watched_events = ["IP_ADDRESS"]
    produced_events = ["GEOLOCATION"]
    flags = ["passive", "safe"]
    meta = {
        "description": "Query IP2location.io's API for geolocation information. ",
        "created_date": "2023-09-12",
        "author": "@TheTechromancer",
        "auth_required": True,
    }
    options = {"api_key": "", "lang": ""}
    options_desc = {
        "api_key": "IP2location.io API Key",
        "lang": "Translation information(ISO639-1). The translation is only applicable for continent, country, region and city name.",
    }
    scope_distance_modifier = 1
    _priority = 2
    suppress_dupes = False

    base_url = "http://api.ip2location.io"

    async def setup(self):
        self.lang = self.config.get("lang", "")
        return await self.require_api_key()

    async def ping(self):
        url = self.build_url("8.8.8.8")
        await super().ping(url)

    def build_url(self, data):
        url = f"{self.base_url}/?key={{api_key}}&ip={data}&format=json&source=bbot"
        if self.lang:
            url = f"{url}&lang={self.lang}"
        return url

    async def handle_event(self, event):
        try:
            url = self.build_url(event.data)
            result = await self.api_request(url)
            if result:
                geo_data = result.json()
                if not geo_data:
                    self.verbose(f"No JSON response from {url}")
            else:
                self.verbose(f"No response from {url}")
        except Exception:
            self.verbose(f"Error retrieving results for {event.data}", trace=True)
            return

        geo_data = {k: v for k, v in geo_data.items() if v is not None}
        if "error" in geo_data:
            error_msg = geo_data.get("error").get("error_message", "")
            if error_msg:
                self.warning(error_msg)
        elif geo_data:
            country = geo_data.get("country_name", "unknown country")
            region = geo_data.get("region_name", "unknown region")
            city = geo_data.get("city_name", "unknown city")
            lat = geo_data.get("latitude", "")
            long = geo_data.get("longitude", "")
            description = f"{city}, {region}, {country} ({lat}, {long})"
            await self.emit_event(
                geo_data,
                "GEOLOCATION",
                event,
                context=f'{{module}} queried IP2Location API for "{event.data}" and found {{event.type}}: {description}',
            )



import re
import asyncio
from werkzeug.wrappers import Response

from .base import ModuleTestBase


def extract_subdomain_tag(data):
    pattern = r"http://([a-z0-9]{4})\.fakedomain\.fakeinteractsh\.com"
    match = re.search(pattern, data)
    if match:
        return match.group(1)


class TestGeneric_SSRF(ModuleTestBase):
    targets = ["http://127.0.0.1:8888"]
    modules_overrides = ["httpx", "generic_ssrf"]

    def request_handler(self, request):
        subdomain_tag = None

        if request.method == "GET":
            subdomain_tag = extract_subdomain_tag(request.full_path)
        elif request.method == "POST":
            subdomain_tag = extract_subdomain_tag(request.data.decode())
        if subdomain_tag:
            asyncio.run(
                self.interactsh_mock_instance.mock_interaction(
                    subdomain_tag, msg=f"{request.method}: {request.data.decode()}"
                )
            )

        return Response("alive", status=200)

    async def setup_before_prep(self, module_test):
        self.interactsh_mock_instance = module_test.mock_interactsh("generic_ssrf")
        module_test.monkeypatch.setattr(
            module_test.scan.helpers, "interactsh", lambda *args, **kwargs: self.interactsh_mock_instance
        )

    async def setup_after_prep(self, module_test):
        expect_args = re.compile("/")
        module_test.set_expect_requests_handler(expect_args=expect_args, request_handler=self.request_handler)

    def check(self, module_test, events):
        total_vulnerabilities = 0
        total_findings = 0

        for e in events:
            if e.type == "VULNERABILITY":
                total_vulnerabilities += 1
            elif e.type == "FINDING":
                total_findings += 1

        assert total_vulnerabilities == 30, "Incorrect number of vulnerabilities detected"
        assert total_findings == 30, "Incorrect number of findings detected"

        assert any(
            e.type == "VULNERABILITY"
            and "Out-of-band interaction: [Generic SSRF (GET)]"
            and "[Triggering Parameter: Dest]" in e.data["description"]
            for e in events
        ), "Failed to detect Generic SSRF (GET)"
        assert any(
            e.type == "VULNERABILITY" and "Out-of-band interaction: [Generic SSRF (POST)]" in e.data["description"]
            for e in events
        ), "Failed to detect Generic SSRF (POST)"
        assert any(
            e.type == "VULNERABILITY" and "Out-of-band interaction: [Generic XXE] [HTTP]" in e.data["description"]
            for e in events
        ), "Failed to detect Generic SSRF (XXE)"


class TestGeneric_SSRF_httponly(TestGeneric_SSRF):
    config_overrides = {"modules": {"generic_ssrf": {"skip_dns_interaction": True}}}

    def check(self, module_test, events):
        total_vulnerabilities = 0
        total_findings = 0

        for e in events:
            if e.type == "VULNERABILITY":
                total_vulnerabilities += 1
            elif e.type == "FINDING":
                total_findings += 1

        assert total_vulnerabilities == 30, "Incorrect number of vulnerabilities detected"
        assert total_findings == 0, "Incorrect number of findings detected"



from .test_module_paramminer_headers import Paramminer_Headers, tempwordlist, helper


class TestParamminer_Getparams(Paramminer_Headers):
    modules_overrides = ["httpx", "paramminer_getparams"]
    config_overrides = {"modules": {"paramminer_getparams": {"wordlist": tempwordlist(["canary", "id"])}}}

    getparam_body = """
    <html>
    <title>the title</title>
    <body>
    <p>Hello null!</p>';
    </body>
    </html>
    """

    getparam_body_match = """
    <html>
    <title>the title</title>
    <body>
    <p>Hello AAAAAAAAAAAAAA!</p>';
    </body>
    </html>
    """

    async def setup_after_prep(self, module_test):
        module_test.scan.modules["paramminer_getparams"].rand_string = lambda *args, **kwargs: "AAAAAAAAAAAAAA"
        module_test.monkeypatch.setattr(
            helper.HttpCompare, "gen_cache_buster", lambda *args, **kwargs: {"AAAAAA": "1"}
        )
        expect_args = {"query_string": b"id=AAAAAAAAAAAAAA&AAAAAA=1"}
        respond_args = {"response_data": self.getparam_body_match}
        module_test.set_expect_requests(expect_args=expect_args, respond_args=respond_args)

        respond_args = {"response_data": self.getparam_body}
        module_test.set_expect_requests(respond_args=respond_args)

    def check(self, module_test, events):
        assert any(
            e.type == "WEB_PARAMETER"
            and "[Paramminer] Getparam: [id] Reasons: [body] Reflection: [True]" in e.data["description"]
            for e in events
        )
        assert not any(
            e.type == "WEB_PARAMETER" and "[Paramminer] Getparam: [canary] Reasons: [body]" in e.data["description"]
            for e in events
        )


class TestParamminer_Getparams_noreflection(TestParamminer_Getparams):
    getparam_body_match = """
    <html>
    <title>the title</title>
    <body>
    <p>Hello ADMINISTRATOR!</p>';
    </body>
    </html>
    """

    def check(self, module_test, events):
        assert any(
            e.type == "WEB_PARAMETER"
            and "[Paramminer] Getparam: [id] Reasons: [body] Reflection: [False]" in e.data["description"]
            for e in events
        )


class TestParamminer_Getparams_singlewordlist(TestParamminer_Getparams):
    config_overrides = {"modules": {"paramminer_getparams": {"wordlist": tempwordlist(["id"])}}}


class TestParamminer_Getparams_boring_off(TestParamminer_Getparams):
    config_overrides = {
        "modules": {
            "paramminer_getparams": {"skip_boring_words": False, "wordlist": tempwordlist(["canary", "utm_term"])}
        }
    }

    async def setup_after_prep(self, module_test):
        module_test.scan.modules["paramminer_getparams"].rand_string = lambda *args, **kwargs: "AAAAAAAAAAAAAA"
        module_test.monkeypatch.setattr(
            helper.HttpCompare, "gen_cache_buster", lambda *args, **kwargs: {"AAAAAA": "1"}
        )
        expect_args = {"query_string": b"utm_term=AAAAAAAAAAAAAA&AAAAAA=1"}
        respond_args = {"response_data": self.getparam_body_match}
        module_test.set_expect_requests(expect_args=expect_args, respond_args=respond_args)

        respond_args = {"response_data": self.getparam_body}
        module_test.set_expect_requests(respond_args=respond_args)

    def check(self, module_test, events):
        emitted_boring_parameter = False
        for e in events:
            if e.type == "WEB_PARAMETER":
                if "utm_term" in e.data["description"]:
                    emitted_boring_parameter = True
        assert emitted_boring_parameter, "failed to emit boring parameter with skip_boring_words disabled"


class TestParamminer_Getparams_boring_on(TestParamminer_Getparams_boring_off):
    config_overrides = {
        "modules": {
            "paramminer_getparams": {"skip_boring_words": True, "wordlist": tempwordlist(["canary", "boring"])}
        }
    }

    def check(self, module_test, events):
        emitted_boring_parameter = False

        for e in events:
            if e.type == "WEB_PARAMETER":
                if "boring" in e.data["description"]:
                    emitted_boring_parameter = True

        assert not emitted_boring_parameter, "emitted boring parameter with skip_boring_words enabled"


class TestParamminer_Getparams_finish(Paramminer_Headers):
    modules_overrides = ["httpx", "excavate", "paramminer_getparams"]
    config_overrides = {
        "modules": {"paramminer_getparams": {"wordlist": tempwordlist(["canary", "canary2"]), "recycle_words": True}}
    }

    targets = ["http://127.0.0.1:8888/test1.php", "http://127.0.0.1:8888/test2.php"]

    test_1_html = """
<html><a href="/test2.php?abcd1234=foo">paramstest2</a></html>
    """

    test_2_html = """
<html></a><p>Hello</p></html>
    """

    test_2_html_match = """
<html></a><p>HackThePlanet!</p></html>
    """

    async def setup_after_prep(self, module_test):
        module_test.scan.modules["paramminer_getparams"].rand_string = lambda *args, **kwargs: "AAAAAAAAAAAAAA"
        module_test.monkeypatch.setattr(
            helper.HttpCompare, "gen_cache_buster", lambda *args, **kwargs: {"AAAAAA": "1"}
        )

        expect_args = {"uri": "/test2.php", "query_string": b"abcd1234=AAAAAAAAAAAAAA&AAAAAA=1"}
        respond_args = {"response_data": self.test_2_html_match}
        module_test.set_expect_requests(expect_args=expect_args, respond_args=respond_args)

        expect_args = {"uri": "/test2.php"}
        respond_args = {"response_data": self.test_2_html}
        module_test.set_expect_requests(expect_args=expect_args, respond_args=respond_args)

        expect_args = {"uri": "/test1.php", "query_string": b"abcd1234=AAAAAAAAAAAAAA&AAAAAA=1"}
        respond_args = {"response_data": self.test_2_html_match}
        module_test.set_expect_requests(expect_args=expect_args, respond_args=respond_args)

        expect_args = {"uri": "/test1.php"}
        respond_args = {"response_data": self.test_1_html}
        module_test.set_expect_requests(expect_args=expect_args, respond_args=respond_args)

    def check(self, module_test, events):
        excavate_extracted_web_parameter = False
        found_hidden_getparam_recycled = False
        emitted_excavate_paramminer_duplicate = False

        for e in events:
            if e.type == "WEB_PARAMETER":
                if (
                    "http://127.0.0.1:8888/test2.php" in e.data["url"]
                    and "HTTP Extracted Parameter [abcd1234] (HTML Tags Submodule)" in e.data["description"]
                ):
                    excavate_extracted_web_parameter = True

                if (
                    "http://127.0.0.1:8888/test1.php" in e.data["url"]
                    and "[Paramminer] Getparam: [abcd1234] Reasons: [body] Reflection: [False]"
                    in e.data["description"]
                ):
                    found_hidden_getparam_recycled = True

                if (
                    "http://127.0.0.1:8888/test2.php" in e.data["url"]
                    and "[Paramminer] Getparam: [abcd1234] Reasons: [body] Reflection: [False]"
                    in e.data["description"]
                ):
                    emitted_excavate_paramminer_duplicate = True

        assert excavate_extracted_web_parameter, "Excavate failed to extract GET parameter"
        assert found_hidden_getparam_recycled, "Failed to find hidden GET parameter"
        # the fact that it is a duplicate is OK, because it still won't be consumed mutltiple times. But we do want to make sure both modules try to emit it
        assert emitted_excavate_paramminer_duplicate, "Paramminer emitted duplicate already found by excavate"


class TestParamminer_Getparams_xmlspeculative(Paramminer_Headers):
    targets = ["http://127.0.0.1:8888/"]
    modules_overrides = ["httpx", "excavate", "paramminer_getparams"]
    config_overrides = {
        "modules": {"paramminer_getparams": {"wordlist": tempwordlist(["data", "common"]), "recycle_words": False}}
    }
    getparam_extract_xml = """
    <data>
     <junkparameter>1</junkparameter>
     <obscureParameter>1</obscureParameter>
         <common>1</common>
     </data>
    """

    getparam_speculative_used = """
    <html>
    <p>secret parameter used</p>
    </html>
    """

    async def setup_after_prep(self, module_test):
        module_test.scan.modules["paramminer_getparams"].rand_string = lambda *args, **kwargs: "AAAAAAAAAAAAAA"
        module_test.monkeypatch.setattr(
            helper.HttpCompare, "gen_cache_buster", lambda *args, **kwargs: {"AAAAAA": "1"}
        )
        expect_args = {"query_string": b"obscureParameter=AAAAAAAAAAAAAA&AAAAAA=1"}
        respond_args = {"response_data": self.getparam_speculative_used}
        module_test.set_expect_requests(expect_args=expect_args, respond_args=respond_args)

        expect_args = {"query_string": b"data=AAAAAAAAAAAAAA&obscureParameter=AAAAAAAAAAAAAA&AAAAAA=1"}
        respond_args = {"response_data": self.getparam_speculative_used}
        module_test.set_expect_requests(expect_args=expect_args, respond_args=respond_args)

        respond_args = {"response_data": self.getparam_extract_xml, "headers": {"Content-Type": "application/xml"}}
        module_test.set_expect_requests(respond_args=respond_args)

    def check(self, module_test, events):
        excavate_discovered_speculative = False
        paramminer_used_speculative = False
        for e in events:
            if e.type == "WEB_PARAMETER":
                if (
                    "HTTP Extracted Parameter (speculative from xml content) [obscureParameter]"
                    in e.data["description"]
                ):
                    excavate_discovered_speculative = True

                if (
                    "[Paramminer] Getparam: [obscureParameter] Reasons: [header,body] Reflection: [False]"
                    in e.data["description"]
                ):
                    paramminer_used_speculative = True

        assert excavate_discovered_speculative, "Excavate failed to discover speculative xml parameter"
        assert paramminer_used_speculative, "Paramminer failed to confirm speculative GET parameter"


class TestParamminer_Getparams_filter_static(TestParamminer_Getparams_finish):
    targets = ["http://127.0.0.1:8888/test1.php", "http://127.0.0.1:8888/test2.pdf"]

    test_1_html = """
    <html><a href="/test2.pdf?abcd1234=foo">paramstest2</a></html>
    """

    def check(self, module_test, events):
        found_hidden_getparam_recycled = False
        emitted_excavate_paramminer_duplicate = False

        for e in events:
            if e.type == "WEB_PARAMETER":
                if (
                    "http://127.0.0.1:8888/test1.php" in e.data["url"]
                    and "[Paramminer] Getparam: [abcd1234] Reasons: [body] Reflection: [False]"
                    in e.data["description"]
                ):
                    found_hidden_getparam_recycled = True

                if (
                    "http://127.0.0.1:8888/test2.pdf" in e.data["url"]
                    and "[Paramminer] Getparam: [abcd1234] Reasons: [body] Reflection: [False]"
                    in e.data["description"]
                ):
                    emitted_excavate_paramminer_duplicate = True

        assert found_hidden_getparam_recycled, "Failed to find hidden GET parameter"
        assert not emitted_excavate_paramminer_duplicate, "Paramminer emitted parameter for static URL"



import zipfile
import json
from pathlib import Path
from bbot.modules.templates.postman import postman


class postman_download(postman):
    watched_events = ["CODE_REPOSITORY"]
    produced_events = ["FILESYSTEM"]
    flags = ["passive", "subdomain-enum", "safe", "code-enum"]
    meta = {
        "description": "Download workspaces, collections, requests from Postman",
        "created_date": "2024-09-07",
        "author": "@domwhewell-sage",
    }
    options = {"output_folder": "", "api_key": ""}
    options_desc = {"output_folder": "Folder to download postman workspaces to", "api_key": "Postman API Key"}
    scope_distance_modifier = 2

    async def setup(self):
        output_folder = self.config.get("output_folder")
        if output_folder:
            self.output_dir = Path(output_folder) / "postman_workspaces"
        else:
            self.output_dir = self.scan.home / "postman_workspaces"
        self.helpers.mkdir(self.output_dir)
        return await super().setup()

    async def filter_event(self, event):
        if event.type == "CODE_REPOSITORY":
            if "postman" not in event.tags:
                return False, "event is not a postman workspace"
        return True

    async def handle_event(self, event):
        repo_url = event.data.get("url")
        workspace_id = await self.get_workspace_id(repo_url)
        if workspace_id:
            self.verbose(f"Found workspace ID {workspace_id} for {repo_url}")
            data = await self.request_workspace(workspace_id)
            workspace = data["workspace"]
            environments = data["environments"]
            collections = data["collections"]
            workspace_path = self.save_workspace(workspace, environments, collections)
            if workspace_path:
                self.verbose(f"Downloaded workspace from {repo_url} to {workspace_path}")
                codebase_event = self.make_event(
                    {"path": str(workspace_path)}, "FILESYSTEM", tags=["postman", "workspace"], parent=event
                )
                await self.emit_event(
                    codebase_event,
                    context=f"{{module}} downloaded postman workspace at {repo_url} to {{event.type}}: {workspace_path}",
                )

    def save_workspace(self, workspace, environments, collections):
        zip_path = None
        # Create a folder for the workspace
        name = workspace["name"]
        id = workspace["id"]
        folder = self.output_dir / name
        self.helpers.mkdir(folder)
        zip_path = folder / f"{id}.zip"

        # Main Workspace
        self.add_json_to_zip(zip_path, workspace, f"{name}.postman_workspace.json")

        # Workspace Environments
        if environments:
            for environment in environments:
                environment_id = environment["id"]
                self.add_json_to_zip(zip_path, environment, f"{environment_id}.postman_environment.json")

            # Workspace Collections
            if collections:
                for collection in collections:
                    collection_name = collection["info"]["name"]
                    self.add_json_to_zip(zip_path, collection, f"{collection_name}.postman_collection.json")
        return zip_path

    def add_json_to_zip(self, zip_path, data, filename):
        with zipfile.ZipFile(zip_path, "a") as zipf:
            json_content = json.dumps(data, indent=4)
            zipf.writestr(filename, json_content)



from lxml import etree
from bbot.modules.base import BaseModule


class dastardly(BaseModule):
    watched_events = ["HTTP_RESPONSE"]
    produced_events = ["FINDING", "VULNERABILITY"]
    flags = ["active", "aggressive", "slow", "web-thorough"]
    meta = {
        "description": "Lightweight web application security scanner",
        "created_date": "2023-12-11",
        "author": "@domwhewell-sage",
    }

    deps_pip = ["lxml~=5.3.0"]
    deps_common = ["docker"]
    per_hostport_only = True

    default_discovery_context = "{module} performed a light web scan against {event.parent.data['url']} and discovered {event.data['description']} at {event.data['url']}"

    async def setup(self):
        await self.run_process("systemctl", "start", "docker", sudo=True)
        await self.run_process("docker", "pull", "public.ecr.aws/portswigger/dastardly:latest", sudo=True)
        self.output_dir = self.scan.home / "dastardly"
        self.helpers.mkdir(self.output_dir)
        return True

    async def filter_event(self, event):
        # Reject redirects. This helps to avoid scanning the same site twice.
        is_redirect = str(event.data["status_code"]).startswith("30")
        if is_redirect:
            return False, "URL is a redirect"
        return True

    async def handle_event(self, event):
        host = event.parsed_url._replace(path="/").geturl()
        self.verbose(f"Running Dastardly scan against {host}")
        command, output_file = self.construct_command(host)
        finished_proc = await self.run_process(command, sudo=True)
        self.debug(f"dastardly stdout: {getattr(finished_proc, 'stdout', '')}")
        self.debug(f"dastardly stderr: {getattr(finished_proc, 'stderr', '')}")
        for testsuite in self.parse_dastardly_xml(output_file):
            url = testsuite.endpoint
            for testcase in testsuite.testcases:
                for failure in testcase.failures:
                    if failure.severity == "Info":
                        await self.emit_event(
                            {
                                "host": str(event.host),
                                "url": url,
                                "description": failure.instance,
                            },
                            "FINDING",
                            event,
                            context=f"{{module}} executed web scan against {host} and identified {{event.type}}: {failure.instance}",
                        )
                    else:
                        await self.emit_event(
                            {
                                "severity": failure.severity,
                                "host": str(event.host),
                                "url": url,
                                "description": failure.instance,
                            },
                            "VULNERABILITY",
                            event,
                            context=f"{{module}} executed web scan against {host} and identified {failure.severity.lower()} {{event.type}}: {failure.instance}",
                        )

    def construct_command(self, target):
        date_time = self.helpers.make_date()
        file_name = self.helpers.tagify(target)
        temp_path = self.output_dir / f"{date_time}_{file_name}.xml"
        command = [
            "docker",
            "run",
            "--user",
            "0",
            "--rm",
            "-v",
            f"{self.output_dir}:/dastardly",
            "-e",
            f"BURP_START_URL={target}",
            "-e",
            f"BURP_REPORT_FILE_PATH=/dastardly/{temp_path.name}",
            "public.ecr.aws/portswigger/dastardly:latest",
        ]
        return command, temp_path

    def parse_dastardly_xml(self, xml_file):
        try:
            with open(xml_file, "rb") as f:
                et = etree.parse(f, parser=etree.XMLParser(recover=True, resolve_entities=False))
                for testsuite in et.iter("testsuite"):
                    yield TestSuite(testsuite)
        except FileNotFoundError:
            self.debug(f"Could not find Dastardly XML file at {xml_file}")
        except OSError as e:
            self.verbose(f"Error opening Dastardly XML file at {xml_file}: {e}")
        except etree.ParseError as e:
            self.warning(f"Error parsing Dastardly XML at {xml_file}: {e}")


class Failure:
    def __init__(self, xml):
        self.etree = xml

        # instance information
        self.instance = self.etree.attrib.get("message", "")
        self.severity = self.etree.attrib.get("type", "")
        self.text = self.etree.text


class TestCase:
    def __init__(self, xml):
        self.etree = xml

        # title information
        self.title = self.etree.attrib.get("name", "")

        # findings / failures(as dastardly names them)
        self.failures = []
        for failure in self.etree.findall("failure"):
            self.failures.append(Failure(failure))


class TestSuite:
    def __init__(self, xml):
        self.etree = xml

        # endpoint information
        self.endpoint = self.etree.attrib.get("name", "")

        # test cases
        self.testcases = []
        for testcase in self.etree.findall("testcase"):
            self.testcases.append(TestCase(testcase))



import regex as re
from pathlib import Path
from subprocess import CalledProcessError
from bbot.modules.base import BaseModule


class gitdumper(BaseModule):
    watched_events = ["CODE_REPOSITORY"]
    produced_events = ["FILESYSTEM"]
    flags = ["passive", "safe", "slow", "code-enum"]
    meta = {
        "description": "Download a leaked .git folder recursively or by fuzzing common names",
        "created_date": "2025-02-11",
        "author": "@domwhewell-sage",
    }
    options = {
        "output_folder": "",
        "fuzz_tags": False,
        "max_semanic_version": 10,
    }
    options_desc = {
        "output_folder": "Folder to download repositories to",
        "fuzz_tags": "Fuzz for common git tag names (v0.0.1, 0.0.2, etc.) up to the max_semanic_version",
        "max_semanic_version": "Maximum version number to fuzz for (default < v10.10.10)",
    }

    scope_distance_modifier = 2

    async def setup(self):
        self.urls_downloaded = set()
        output_folder = self.config.get("output_folder")
        if output_folder:
            self.output_dir = Path(output_folder) / "git_repos"
        else:
            self.output_dir = self.scan.home / "git_repos"
        self.helpers.mkdir(self.output_dir)
        self.unsafe_regex = self.helpers.re.compile(r"^\s*fsmonitor|sshcommand|askpass|editor|pager", re.IGNORECASE)
        self.ref_regex = self.helpers.re.compile(r"ref: refs/heads/([a-zA-Z\d_-]+)")
        self.obj_regex = self.helpers.re.compile(r"[a-f0-9]{40}")
        self.pack_regex = self.helpers.re.compile(r"pack-([a-f0-9]{40})\.pack")
        self.git_files = [
            "HEAD",
            "description",
            "config",
            "COMMIT_EDITMSG",
            "index",
            "packed-refs",
            "info/refs",
            "info/exclude",
            "refs/stash",
            "refs/wip/index/refs/heads/master",
            "refs/wip/wtree/refs/heads/master",
            "logs/HEAD",
            "objects/info/packs",
        ]
        self.info("Compiling fuzz list with common branch names")
        branch_names = [
            "bugfix",
            "daily",
            "dev",
            "develop",
            "development",
            "feat",
            "feature",
            "fix",
            "hotfix",
            "integration",
            "issue",
            "main",
            "master",
            "ng",
            "prod",
            "production",
            "qa",
            "quickfix",
            "release",
            "stable",
            "stage",
            "staging",
            "test",
            "testing",
            "wip",
        ]
        url_patterns = [
            "logs/refs/heads/{branch}",
            "logs/refs/remotes/origin/{branch}",
            "refs/remotes/origin/{branch}",
            "refs/heads/{branch}",
        ]
        for branch in branch_names:
            for pattern in url_patterns:
                self.git_files.append(pattern.format(branch=branch))
        self.fuzz_tags = self.config.get("fuzz_tags", "10")
        self.max_semanic_version = self.config.get("max_semanic_version", "10")
        if self.fuzz_tags:
            self.info("Adding symantec version tags to fuzz list")
            for major in range(self.max_semanic_version):
                for minor in range(self.max_semanic_version):
                    for patch in range(self.max_semanic_version):
                        self.verbose(f"{major}.{minor}.{patch}")
                        self.git_files.append(f"refs/tags/{major}.{minor}.{patch}")
                        self.verbose(f"v{major}.{minor}.{patch}")
                        self.git_files.append(f"refs/tags/v{major}.{minor}.{patch}")
        else:
            self.info("Adding symantec version tags to fuzz list (v0.0.1, 0.0.1, v1.0.0, 1.0.0)")
            for path in ["refs/tags/v0.0.1", "refs/tags/0.0.1", "refs/tags/v1.0.0", "refs/tags/1.0.0"]:
                self.git_files.append(path)
        return await super().setup()

    async def filter_event(self, event):
        if event.type == "CODE_REPOSITORY":
            if "git-directory" not in event.tags:
                return False, "event is not a leaked .git directory"
        return True

    async def handle_event(self, event):
        repo_url = event.data.get("url")
        self.info(f"Processing leaked .git directory at {repo_url}")
        repo_folder = self.output_dir / self.helpers.tagify(repo_url)
        self.helpers.mkdir(repo_folder)
        dir_listing = await self.directory_listing_enabled(repo_url)
        if dir_listing:
            urls = await self.recursive_dir_list(dir_listing)
            result = await self.download_files(urls, repo_folder)
        else:
            result = await self.git_fuzz(repo_url, repo_folder)
        if result:
            await self.sanitize_config(repo_folder)
            await self.git_checkout(repo_folder)
            codebase_event = self.make_event({"path": str(repo_folder)}, "FILESYSTEM", tags=["git"], parent=event)
            await self.emit_event(
                codebase_event,
                context=f"{{module}} cloned git repo at {repo_url} to {{event.type}}: {str(repo_folder)}",
            )
        else:
            self.helpers.rm_rf(repo_folder)

    async def directory_listing_enabled(self, repo_url):
        response = await self.helpers.request(repo_url)
        if "<title>Index of" in getattr(response, "text", ""):
            self.info(f"Directory listing enabled at {repo_url}")
            return response
        return None

    async def recursive_dir_list(self, dir_listing):
        file_list = []
        soup = self.helpers.beautifulsoup(dir_listing.text, "html.parser")
        links = soup.find_all("a")
        for link in links:
            href = link["href"]
            if href == "../" or href == "/":
                continue
            if href.endswith("/"):
                folder_url = self.helpers.urljoin(str(dir_listing.url), href)
                response = await self.helpers.request(folder_url)
                if getattr(response, "status_code", 0) == 200:
                    file_list.extend(await self.recursive_dir_list(response))
            else:
                file_url = self.helpers.urljoin(str(dir_listing.url), href)
                # Ensure the file is in the same domain as the directory listing
                if file_url.startswith(str(dir_listing.url)):
                    url = self.helpers.urlparse(file_url)
                    file_list.append(url)
        return file_list

    async def git_fuzz(self, repo_url, repo_folder):
        self.info("Directory listing not enabled, fuzzing common git files")
        url_list = []
        for file in self.git_files:
            url_list.append(self.helpers.urlparse(self.helpers.urljoin(repo_url, file)))
        result = await self.download_files(url_list, repo_folder)
        if result:
            await self.download_current_branch(repo_url, repo_folder)
            await self.download_git_objects(repo_url, repo_folder)
            await self.download_git_packs(repo_url, repo_folder)
            return True
        else:
            return False

    async def download_current_branch(self, repo_url, repo_folder):
        for branch in await self.regex_files(self.ref_regex, file=repo_folder / ".git/HEAD"):
            await self.download_files(
                [self.helpers.urlparse(self.helpers.urljoin(repo_url, f"refs/heads/{branch}"))], repo_folder
            )

    async def download_git_objects(self, url, folder):
        for object in await self.regex_files(self.obj_regex, folder=folder):
            await self.download_object(object, url, folder)

    async def download_git_packs(self, url, folder):
        url_list = []
        for sha1 in await self.regex_files(self.pack_regex, file=folder / ".git/objects/info/packs"):
            url_list.append(self.helpers.urlparse(self.helpers.urljoin(url, f"objects/pack/pack-{sha1}.idx")))
            url_list.append(self.helpers.urlparse(self.helpers.urljoin(url, f"objects/pack/pack-{sha1}.pack")))
        if url_list:
            await self.download_files(url_list, folder)

    async def regex_files(self, regex, folder=Path(), file=Path(), files=[]):
        results = []
        if folder:
            if folder.is_dir():
                for file_path in folder.rglob("*"):
                    if file_path.is_file():
                        results.extend(await self.regex_file(regex, file_path))
        if files:
            for file in files:
                results.extend(await self.regex_file(regex, file))
        if file:
            results.extend(await self.regex_file(regex, file))
        return results

    async def regex_file(self, regex, file=Path()):
        if file.exists() and file.is_file():
            with file.open("r", encoding="utf-8", errors="ignore") as file:
                content = file.read()
                matches = await self.helpers.re.findall(regex, content)
                if matches:
                    return matches
        return []

    async def download_object(self, object, repo_url, repo_folder):
        await self.download_files(
            [self.helpers.urlparse(self.helpers.urljoin(repo_url, f"objects/{object[:2]}/{object[2:]}"))], repo_folder
        )
        output = await self.git_catfile(object, option="-p", folder=repo_folder)
        for obj in await self.helpers.re.findall(self.obj_regex, output):
            await self.download_object(obj, repo_url, repo_folder)

    async def download_files(self, urls, folder):
        for url in urls:
            git_index = url.path.find(".git")
            file_url = url.geturl()
            filename = folder / url.path[git_index:]
            self.helpers.mkdir(filename.parent)
            if hash(str(file_url)) not in self.urls_downloaded:
                self.verbose(f"Downloading {file_url} to {filename}")
                await self.helpers.download(file_url, filename=filename, warn=False)
                self.urls_downloaded.add(hash(str(file_url)))
        if any(folder.rglob("*")):
            return True
        else:
            self.debug(f"Unable to download git files to {folder}")
            return False

    async def sanitize_config(self, folder):
        config_file = folder / ".git/config"
        if config_file.exists():
            with config_file.open("r", encoding="utf-8", errors="ignore") as file:
                content = file.read()
                sanitized = await self.helpers.re.sub(self.unsafe_regex, r"# \g<0>", content)
            with config_file.open("w", encoding="utf-8") as file:
                file.write(sanitized)

    async def git_catfile(self, hash, option="-t", folder=Path()):
        command = ["git", "cat-file", option, hash]
        try:
            output = await self.run_process(command, env={"GIT_TERMINAL_PROMPT": "0"}, cwd=folder, check=True)
        except CalledProcessError:
            return ""

        return output.stdout

    async def git_checkout(self, folder):
        self.verbose(f"Running git checkout to reconstruct the git repository at {folder}")
        command = ["git", "checkout", "."]
        try:
            await self.run_process(command, env={"GIT_TERMINAL_PROMPT": "0"}, cwd=folder, check=True)
        except CalledProcessError as e:
            # Still emit the event even if the checkout fails
            self.debug(f"Error running git checkout in {folder}. STDERR: {repr(e.stderr)}")



from pathlib import Path
from bbot.modules.internal.base import BaseInternalModule
from bbot.core.helpers.libmagic import get_magic_info, get_compression


class unarchive(BaseInternalModule):
    watched_events = ["FILESYSTEM"]
    produced_events = ["FILESYSTEM"]
    flags = ["passive", "safe"]
    meta = {
        "description": "Extract different types of files into folders on the filesystem",
        "created_date": "2024-12-08",
        "author": "@domwhewell-sage",
    }

    async def setup(self):
        self.ignore_compressions = ["application/java-archive", "application/vnd.android.package-archive"]
        self.compression_methods = {
            "zip": ["7z", "x", '-p""', "-aoa", "{filename}", "-o{extract_dir}/"],
            "bzip2": ["tar", "--overwrite", "-xvjf", "{filename}", "-C", "{extract_dir}/"],
            "xz": ["tar", "--overwrite", "-xvJf", "{filename}", "-C", "{extract_dir}/"],
            "7z": ["7z", "x", '-p""', "-aoa", "{filename}", "-o{extract_dir}/"],
            # "rar": ["7z", "x", '-p""', "-aoa", "{filename}", "-o{extract_dir}/"],
            # "lzma": ["7z", "x", '-p""', "-aoa", "{filename}", "-o{extract_dir}/"],
            "tar": ["tar", "--overwrite", "-xvf", "{filename}", "-C", "{extract_dir}/"],
            "gzip": ["tar", "--overwrite", "-xvzf", "{filename}", "-C", "{extract_dir}/"],
        }
        return True

    async def filter_event(self, event):
        if "file" in event.tags:
            magic_mime_type = event.data.get("magic_mime_type", "")
            if magic_mime_type in self.ignore_compressions:
                return False, f"Ignoring file type: {magic_mime_type}, {event.data['path']}"
            if "compression" in event.data:
                if not event.data["compression"] in self.compression_methods:
                    return (
                        False,
                        f"Extract unable to handle file type: {event.data['compression']}, {event.data['path']}",
                    )
            else:
                return False, f"Event is not a compressed file: {event.data['path']}"
        else:
            return False, "Event is not a file"
        return True

    async def handle_event(self, event):
        path = Path(event.data["path"])
        output_dir = path.parent / path.name.replace(".", "_")

        # Use the appropriate extraction method based on the file type
        self.info(f"Extracting {path} to {output_dir}")
        success = await self.extract_file(path, output_dir)

        # If the extraction was successful, emit the event
        if success:
            await self.emit_event(
                {"path": str(output_dir)},
                "FILESYSTEM",
                tags=["folder", "unarchived-folder"],
                parent=event,
                context=f'extracted "{path}" to: {output_dir}',
            )
        else:
            output_dir.rmdir()

    async def extract_file(self, path, output_dir):
        extension, mime_type, description, confidence = get_magic_info(path)
        compression_format = get_compression(mime_type)
        cmd_list = self.compression_methods.get(compression_format, [])
        if cmd_list:
            if not output_dir.exists():
                self.helpers.mkdir(output_dir)
            command = [s.format(filename=path, extract_dir=output_dir) for s in cmd_list]
            try:
                await self.run_process(command, check=True)
                for item in output_dir.iterdir():
                    if item.is_file():
                        await self.extract_file(item, output_dir / item.stem)
            except Exception as e:
                self.warning(f"Error extracting {path}. Error: {e}")
                return False
            return True



from bbot.errors import NTLMError
from bbot.modules.base import BaseModule

ntlm_discovery_endpoints = [
    "",
    "autodiscover/autodiscover.xml",
    "ecp/",
    "ews/",
    "ews/exchange.asmx",
    "exchange/",
    "exchweb/",
    "oab/",
    "owa/",
    "_windows/default.aspx?ReturnUrl=/",
    "abs/",
    "adfs/ls/wia",
    "adfs/services/trust/2005/windowstransport",
    "aspnet_client/",
    "autodiscover/",
    "autoupdate/",
    "certenroll/",
    "certprov/",
    "certsrv/",
    "conf/",
    "debug/",
    "deviceupdatefiles_ext/",
    "deviceupdatefiles_int/",
    "dialin/",
    "etc/",
    "groupexpansion/",
    "hybridconfig/",
    "iwa/authenticated.aspx",
    "iwa/iwa_test.aspx",
    "mcx/",
    "mcx/mcxservice.svc",
    "meet/",
    "meeting/",
    "microsoft-server-activesync/",
    "ocsp/",
    "persistentchat/",
    "phoneconferencing/",
    "powershell/",
    "public/",
    "reach/sip.svc",
    "remoteDesktopGateway/",
    "requesthandler/",
    "requesthandlerext/",
    "rgs/",
    "rgsclients/",
    "rpc/",
    "rpcwithcert/",
    "scheduler/",
    "ucwa/",
    "unifiedmessaging/",
    "webticket/",
    "webticket/webticketservice.svc",
]

NTLM_test_header = {"Authorization": "NTLM TlRMTVNTUAABAAAAl4II4gAAAAAAAAAAAAAAAAAAAAAKAGFKAAAADw=="}


class ntlm(BaseModule):
    """
    Todo:
        Cancel pending requests and break out of loop when valid endpoint is found
        (waiting on https://github.com/encode/httpcore/discussions/783/ to be fixed first)
    """

    watched_events = ["URL", "HTTP_RESPONSE"]
    produced_events = ["FINDING", "DNS_NAME"]
    flags = ["active", "safe", "web-basic"]
    meta = {
        "description": "Watch for HTTP endpoints that support NTLM authentication",
        "created_date": "2022-07-25",
        "author": "@liquidsec",
    }
    options = {"try_all": False}
    options_desc = {"try_all": "Try every NTLM endpoint"}

    in_scope_only = True

    async def setup(self):
        self.found = set()
        self.try_all = self.config.get("try_all", False)
        return True

    async def handle_event(self, event):
        found_hash = hash(f"{event.host}:{event.port}")
        if event.type == "URL":
            url = event.data
        else:
            url = event.data["url"]
        if found_hash in self.found:
            return

        urls = {url}
        if self.try_all:
            for endpoint in ntlm_discovery_endpoints:
                urls.add(f"{event.parsed_url.scheme}://{event.parsed_url.netloc}/{endpoint}")

        num_urls = len(urls)
        agen = self.helpers.request_batch(
            urls, headers=NTLM_test_header, allow_redirects=False, timeout=self.http_timeout
        )
        async for url, response in agen:
            ntlm_resp = response.headers.get("WWW-Authenticate", "")
            if not ntlm_resp:
                continue
            ntlm_resp_b64 = max(ntlm_resp.split(","), key=lambda x: len(x)).split()[-1]
            try:
                ntlm_resp_decoded = self.helpers.ntlm.ntlmdecode(ntlm_resp_b64)
                if not ntlm_resp_decoded:
                    continue

                await agen.aclose()
                self.found.add(found_hash)
                fqdn = ntlm_resp_decoded.get("FQDN", "")
                await self.emit_event(
                    {
                        "host": str(event.host),
                        "url": url,
                        "description": f"NTLM AUTH: {ntlm_resp_decoded}",
                    },
                    "FINDING",
                    parent=event,
                    context=f"{{module}} tried {num_urls:,} NTLM endpoints against {url} and identified NTLM auth ({{event.type}}): {fqdn}",
                )
                fqdn = ntlm_resp_decoded.get("FQDN", "")
                if fqdn:
                    await self.emit_event(fqdn, "DNS_NAME", parent=event)
                break

            except NTLMError as e:
                self.verbose(str(e))

    async def filter_event(self, event):
        if self.try_all:
            return True
        if event.type == "HTTP_RESPONSE":
            if "www-authenticate" in event.data["header-dict"]:
                header_value = event.data["header-dict"]["www-authenticate"][0].lower()
                if "ntlm" in header_value or "negotiate" in header_value:
                    return True
        return False



import asyncio

from pathlib import Path
from .base import ModuleTestBase


class TestUnarchive(ModuleTestBase):
    targets = ["http://127.0.0.1:8888"]
    modules_overrides = ["filedownload", "httpx", "excavate", "speculate", "unarchive"]

    async def setup_after_prep(self, module_test):
        temp_path = Path("/tmp/.bbot_test")

        # Create a text file to compress
        text_file = temp_path / "test.txt"
        with open(text_file, "w") as f:
            f.write("This is a test file")
        zip_file = temp_path / "test.zip"
        zip_zip_file = temp_path / "test_zip.zip"
        bz2_file = temp_path / "test.bz2"
        xz_file = temp_path / "test.xz"
        zip7_file = temp_path / "test.7z"
        # lzma_file = temp_path / "test.lzma"
        tar_file = temp_path / "test.tar"
        tgz_file = temp_path / "test.tgz"
        commands = [
            ("7z", "a", '-p""', "-aoa", f"{zip_file}", f"{text_file}"),
            ("7z", "a", '-p""', "-aoa", f"{zip_zip_file}", f"{zip_file}"),
            ("tar", "-C", f"{temp_path}", "-cvjf", f"{bz2_file}", f"{text_file.name}"),
            ("tar", "-C", f"{temp_path}", "-cvJf", f"{xz_file}", f"{text_file.name}"),
            ("7z", "a", '-p""', "-aoa", f"{zip7_file}", f"{text_file}"),
            # ("tar", "-C", f"{temp_path}", "--lzma", "-cvf", f"{lzma_file}", f"{text_file.name}"),
            ("tar", "-C", f"{temp_path}", "-cvf", f"{tar_file}", f"{text_file.name}"),
            ("tar", "-C", f"{temp_path}", "-cvzf", f"{tgz_file}", f"{text_file.name}"),
        ]

        for command in commands:
            process = await asyncio.create_subprocess_exec(
                *command, stdout=asyncio.subprocess.PIPE, stderr=asyncio.subprocess.PIPE
            )
            stdout, stderr = await process.communicate()
            assert process.returncode == 0, f"Command {command} failed with error: {stderr.decode()}"

        module_test.set_expect_requests(
            dict(uri="/"),
            dict(
                response_data="""<a href="/test.zip">
                <a href="/test-zip.zip">
                <a href="/test.bz2">
                <a href="/test.xz">
                <a href="/test.7z">
                <a href="/test.tar">
                <a href="/test.tgz">""",
            ),
        )
        (
            module_test.set_expect_requests(
                dict(uri="/test.zip"),
                dict(
                    response_data=zip_file.read_bytes(),
                    headers={"Content-Type": "application/zip"},
                ),
            ),
        )
        (
            module_test.set_expect_requests(
                dict(uri="/test-zip.zip"),
                dict(
                    response_data=zip_zip_file.read_bytes(),
                    headers={"Content-Type": "application/zip"},
                ),
            ),
        )
        (
            module_test.set_expect_requests(
                dict(uri="/test.bz2"),
                dict(
                    response_data=bz2_file.read_bytes(),
                    headers={"Content-Type": "application/x-bzip2"},
                ),
            ),
        )
        (
            module_test.set_expect_requests(
                dict(uri="/test.xz"),
                dict(
                    response_data=xz_file.read_bytes(),
                    headers={"Content-Type": "application/x-xz"},
                ),
            ),
        )
        (
            module_test.set_expect_requests(
                dict(uri="/test.7z"),
                dict(
                    response_data=zip7_file.read_bytes(),
                    headers={"Content-Type": "application/x-7z-compressed"},
                ),
            ),
        )
        # (
        #     module_test.set_expect_requests(
        #         dict(uri="/test.rar"),
        #         dict(
        #             response_data=b"Rar!\x1a\x07\x01\x003\x92\xb5\xe5\n\x01\x05\x06\x00\x05\x01\x01\x80\x80\x00\xa2N\x8ec&\x02\x03\x0b\x93\x00\x04\x93\x00\xa4\x83\x02\xc9\x11f\x06\x80\x00\x01\x08test.txt\n\x03\x13S\x96ug\x96\xf3\x1b\x06This is a test file\x1dwVQ\x03\x05\x04\x00",
        #             headers={"Content-Type": "application/vnd.rar"},
        #         ),
        #     ),
        # )
        # (
        #     module_test.set_expect_requests(
        #         dict(uri="/test.lzma"),
        #         dict(
        #             response_data=lzma_file.read_bytes(),
        #             headers={"Content-Type": "application/x-lzma"},
        #         ),
        #     ),
        # )
        (
            module_test.set_expect_requests(
                dict(uri="/test.tar"),
                dict(
                    response_data=tar_file.read_bytes(),
                    headers={"Content-Type": "application/x-tar"},
                ),
            ),
        )
        (
            module_test.set_expect_requests(
                dict(uri="/test.tgz"),
                dict(
                    response_data=tgz_file.read_bytes(),
                    headers={"Content-Type": "application/x-tgz"},
                ),
            ),
        )

    def check(self, module_test, events):
        filesystem_events = [e for e in events if e.type == "FILESYSTEM"]

        # ZIP
        zip_file_event = [e for e in filesystem_events if "test.zip" in e.data["path"]]
        assert 1 == len(zip_file_event), "No zip file found"
        file = Path(zip_file_event[0].data["path"])
        assert file.is_file(), f"File not found at {file}"
        extract_event = [e for e in filesystem_events if "test_zip" in e.data["path"] and "folder" in e.tags]
        assert 1 == len(extract_event), "Failed to extract zip"
        extract_path = Path(extract_event[0].data["path"]) / "test.txt"
        assert extract_path.is_file(), "Failed to extract the test file"

        # Recursive ZIP
        zip_zip_file_event = [e for e in filesystem_events if "test-zip.zip" in e.data["path"]]
        assert 1 == len(zip_zip_file_event), "No recursive file found"
        file = Path(zip_zip_file_event[0].data["path"])
        assert file.is_file(), f"File not found at {file}"
        extract_event = [e for e in filesystem_events if "test-zip_zip" in e.data["path"] and "folder" in e.tags]
        assert 1 == len(extract_event), "Failed to extract zip"
        extract_path = Path(extract_event[0].data["path"]) / "test" / "test.txt"
        assert extract_path.is_file(), "Failed to extract the test file"

        # BZ2
        bz2_file_event = [e for e in filesystem_events if "test.bz2" in e.data["path"]]
        assert 1 == len(bz2_file_event), "No bz2 file found"
        file = Path(bz2_file_event[0].data["path"])
        assert file.is_file(), f"File not found at {file}"
        extract_event = [e for e in filesystem_events if "test_bz2" in e.data["path"] and "folder" in e.tags]
        assert 1 == len(extract_event), "Failed to extract bz2"
        extract_path = Path(extract_event[0].data["path"]) / "test.txt"
        assert extract_path.is_file(), "Failed to extract the test file"

        # XZ
        xz_file_event = [e for e in filesystem_events if "test.xz" in e.data["path"]]
        assert 1 == len(xz_file_event), "No xz file found"
        file = Path(xz_file_event[0].data["path"])
        assert file.is_file(), f"File not found at {file}"
        extract_event = [e for e in filesystem_events if "test_xz" in e.data["path"] and "folder" in e.tags]
        assert 1 == len(extract_event), "Failed to extract xz"
        extract_path = Path(extract_event[0].data["path"]) / "test.txt"
        assert extract_path.is_file(), "Failed to extract the test file"

        # 7z
        zip7_file_event = [e for e in filesystem_events if "test.7z" in e.data["path"]]
        assert 1 == len(zip7_file_event), "No 7z file found"
        file = Path(zip7_file_event[0].data["path"])
        assert file.is_file(), f"File not found at {file}"
        extract_event = [e for e in filesystem_events if "test_7z" in e.data["path"] and "folder" in e.tags]
        assert 1 == len(extract_event), "Failed to extract 7z"
        extract_path = Path(extract_event[0].data["path"]) / "test.txt"
        assert extract_path.is_file(), "Failed to extract the test file"

        # RAR
        # rar_file_event = [e for e in filesystem_events if "test.rar" in e.data["path"]]
        # assert 1 == len(rar_file_event), "No rar file found"
        # file = Path(rar_file_event[0].data["path"])
        # assert file.is_file(), f"File not found at {file}"
        # extract_event = [e for e in filesystem_events if "test_rar" in e.data["path"] and "folder" in e.tags]
        # assert 1 == len(extract_event), "Failed to extract rar"
        # extract_path = Path(extract_event[0].data["path"]) / "test.txt"
        # assert extract_path.is_file(), list(extract_path.parent.iterdir())

        # LZMA
        # lzma_file_event = [e for e in filesystem_events if "test.lzma" in e.data["path"]]
        # assert 1 == len(lzma_file_event), "No lzma file found"
        # file = Path(lzma_file_event[0].data["path"])
        # assert file.is_file(), f"File not found at {file}"
        # extract_event = [e for e in filesystem_events if "test_lzma" in e.data["path"] and "folder" in e.tags]
        # assert 1 == len(extract_event), "Failed to extract lzma"
        # extract_path = Path(extract_event[0].data["path"]) / "test.txt"
        # assert extract_path.is_file(), "Failed to extract the test file"

        # TAR
        tar_file_event = [e for e in filesystem_events if "test.tar" in e.data["path"]]
        assert 1 == len(tar_file_event), "No tar file found"
        file = Path(tar_file_event[0].data["path"])
        assert file.is_file(), f"File not found at {file}"
        extract_event = [e for e in filesystem_events if "test_tar" in e.data["path"] and "folder" in e.tags]
        assert 1 == len(extract_event), "Failed to extract tar"
        extract_path = Path(extract_event[0].data["path"]) / "test.txt"
        assert extract_path.is_file(), "Failed to extract the test file"

        # TGZ
        tgz_file_event = [e for e in filesystem_events if "test.tgz" in e.data["path"]]
        assert 1 == len(tgz_file_event), "No tgz file found"
        file = Path(tgz_file_event[0].data["path"])
        assert file.is_file(), f"File not found at {file}"
        extract_event = [e for e in filesystem_events if "test_tgz" in e.data["path"] and "folder" in e.tags]
        assert 1 == len(extract_event), "Failed to extract tgz"
        extract_path = Path(extract_event[0].data["path"]) / "test.txt"
        assert extract_path.is_file(), "Failed to extract the test file"



import pickle
import re
import random
import string

from bbot.modules.deadly.ffuf import ffuf


class ffuf_shortnames(ffuf):
    watched_events = ["URL_HINT"]
    produced_events = ["URL_UNVERIFIED"]
    flags = ["aggressive", "active", "iis-shortnames", "web-thorough"]
    meta = {
        "description": "Use ffuf in combination IIS shortnames",
        "created_date": "2022-07-05",
        "author": "@liquidsec",
    }

    options = {
        "wordlist_extensions": "",  # default is defined within setup function
        "max_depth": 1,
        "version": "2.0.0",
        "extensions": "",
        "ignore_redirects": True,
        "find_common_prefixes": False,
        "find_delimiters": True,
        "find_subwords": False,
        "max_predictions": 250,
        "rate": 0,
    }

    options_desc = {
        "wordlist_extensions": "Specify wordlist to use when making extension lists",
        "max_depth": "the maximum directory depth to attempt to solve",
        "version": "ffuf version",
        "extensions": "Optionally include a list of extensions to extend the keyword with (comma separated)",
        "ignore_redirects": "Explicitly ignore redirects (301,302)",
        "find_common_prefixes": "Attempt to automatically detect common prefixes and make additional ffuf runs against them",
        "find_delimiters": "Attempt to detect common delimiters and make additional ffuf runs against them",
        "find_subwords": "Attempt to detect subwords and make additional ffuf runs against them",
        "max_predictions": "The maximum number of predictions to generate per shortname prefix",
        "rate": "Rate of requests per second (default: 0)",
    }

    deps_pip = ["numpy"]
    deps_common = ["ffuf"]
    in_scope_only = True

    supplementary_words = ["html", "ajax", "xml", "json", "api"]

    def generate_templist(self, hint, shortname_type):
        virtual_file = set()  # Use a set to avoid duplicates

        for prediction, score in self.predict(hint, self.max_predictions, model=shortname_type):
            prediction_lower = prediction.lower()  # Convert to lowercase
            self.debug(f"Got prediction: [{prediction_lower}] from prefix [{hint}] with score [{score}]")
            virtual_file.add(prediction_lower)  # Add to set to ensure uniqueness

        virtual_file.add(self.canary.lower())  # Ensure canary is also lowercase
        return self.helpers.tempfile(list(virtual_file), pipe=False), len(virtual_file)

    def predict(self, prefix, n=25, model="endpoint"):
        predictor_name = f"{model}_predictor"
        predictor = getattr(self, predictor_name)
        return predictor.predict(prefix, n)

    @staticmethod
    def find_common_prefixes(strings, minimum_set_length=4):
        prefix_candidates = [s[:i] for s in strings if len(s) == 6 for i in range(3, 6)]
        frequency_dict = {item: prefix_candidates.count(item) for item in prefix_candidates}
        frequency_dict = {k: v for k, v in frequency_dict.items() if v >= minimum_set_length}
        prefix_list = list(set(frequency_dict.keys()))

        found_prefixes = set()
        for prefix in prefix_list:
            prefix_frequency = frequency_dict[prefix]
            is_substring = False

            for k, v in frequency_dict.items():
                if prefix != k:
                    if prefix in k:
                        is_substring = True
            if not is_substring:
                found_prefixes.add(prefix)
            else:
                if prefix_frequency > v and (len(k) - len(prefix) == 1):
                    found_prefixes.add(prefix)
        return list(found_prefixes)

    async def setup(self):
        self.proxy = self.scan.web_config.get("http_proxy", "")
        self.canary = "".join(random.choice(string.ascii_lowercase) for i in range(10))
        wordlist_extensions = self.config.get("wordlist_extensions", "")
        if not wordlist_extensions:
            wordlist_extensions = f"{self.helpers.wordlist_dir}/raft-small-extensions-lowercase_CLEANED.txt"
        self.debug(f"Using [{wordlist_extensions}] for shortname candidate extension list")
        self.wordlist_extensions = await self.helpers.wordlist(wordlist_extensions)
        self.ignore_redirects = self.config.get("ignore_redirects")
        self.max_predictions = self.config.get("max_predictions")
        self.find_subwords = self.config.get("find_subwords")
        self.rate = self.config.get("rate", 0)

        class MinimalWordPredictor:
            def __init__(self):
                self.word_frequencies = {}

            def predict(self, prefix, top_n):
                prefix = prefix.lower()
                matches = [(word, freq) for word, freq in self.word_frequencies.items() if word.startswith(prefix)]

                if not matches:
                    return []

                matches.sort(key=lambda x: x[1], reverse=True)
                matches = matches[:top_n]

                max_freq = matches[0][1]
                return [(word, freq / max_freq) for word, freq in matches]

        class CustomUnpickler(pickle.Unpickler):
            def find_class(self, module, name):
                if name == "MinimalWordPredictor":
                    return MinimalWordPredictor
                return super().find_class(module, name)

        self.info("Loading ffuf_shortnames prediction models, could take a while if not cached")
        endpoint_model = await self.helpers.wordlist(
            "https://raw.githubusercontent.com/blacklanternsecurity/wordpredictor/refs/heads/main/trained_models/endpoints.bin"
        )
        directory_model = await self.helpers.wordlist(
            "https://raw.githubusercontent.com/blacklanternsecurity/wordpredictor/refs/heads/main/trained_models/directories.bin"
        )

        self.debug(f"Loading endpoint model from: {endpoint_model}")
        with open(endpoint_model, "rb") as f:
            unpickler = CustomUnpickler(f)
            self.endpoint_predictor = unpickler.load()

        self.debug(f"Loading directory model from: {directory_model}")
        with open(directory_model, "rb") as f:
            unpickler = CustomUnpickler(f)
            self.directory_predictor = unpickler.load()

        self.subword_list = []
        if self.find_subwords:
            self.debug("Acquiring ffuf_shortnames subword list")
            subwords = await self.helpers.wordlist(
                "https://raw.githubusercontent.com/nltk/nltk_data/refs/heads/gh-pages/packages/corpora/words.zip",
                zip=True,
                zip_filename="words/en",
            )
            with open(subwords, "r") as f:
                subword_list_content = f.readlines()
            self.subword_list = {word.lower().strip() for word in subword_list_content if 3 <= len(word.strip()) <= 5}
            self.debug(f"Created subword_list with {len(self.subword_list)} words")
            self.subword_list = self.subword_list.union(self.supplementary_words)
            self.debug(f"Extended subword_list with supplementary words, total size: {len(self.subword_list)}")

        self.per_host_collection = {}
        self.shortname_to_event = {}

        return True

    def build_extension_list(self, event):
        used_extensions = []
        extension_hint = event.parsed_url.path.rsplit(".", 1)[1].lower().strip()
        if len(extension_hint) == 3:
            with open(self.wordlist_extensions) as f:
                for l in f:
                    l = l.lower().lstrip(".")
                    if l.lower().startswith(extension_hint):
                        used_extensions.append(l.strip())
            return used_extensions
        else:
            return [extension_hint]

    def find_delimiter(self, hint):
        delimiters = ["_", "-"]
        for d in delimiters:
            if d in hint:
                if not hint.startswith(d) and not hint.endswith(d):
                    return d, hint.split(d)[0], hint.split(d)[1]
        return None

    async def filter_event(self, event):
        if "iis-magic-url" in event.tags:
            return False, "iis-magic-url URL_HINTs are not solvable by ffuf_shortnames"
        if event.parent.type != "URL":
            return False, "its parent event is not of type URL"
        return True

    def find_subword(self, word):
        for i in range(len(word), 2, -1):  # Start from full length down to 3 characters
            candidate = word[:i]
            if candidate in self.subword_list:
                leftover = word[i:]
                return candidate, leftover
        return None, word  # No match found, return None and the original word

    async def handle_event(self, event):
        filename_hint = re.sub(r"~\d", "", event.parsed_url.path.rsplit(".", 1)[0].split("/")[-1]).lower()

        if "shortname-endpoint" in event.tags:
            shortname_type = "endpoint"
        elif "shortname-directory" in event.tags:
            shortname_type = "directory"
        else:
            self.error("ffuf_shortnames received URL_HINT without proper 'shortname-' tag")
            return

        host = f"{event.parent.parsed_url.scheme}://{event.parent.parsed_url.netloc}/"
        if host not in self.per_host_collection.keys():
            self.per_host_collection[host] = [(filename_hint, event.parent.data)]

        else:
            self.per_host_collection[host].append((filename_hint, event.parent.data))

        self.shortname_to_event[filename_hint] = event

        root_stub = "/".join(event.parsed_url.path.split("/")[:-1])
        root_url = f"{event.parsed_url.scheme}://{event.parsed_url.netloc}{root_stub}/"

        if shortname_type == "endpoint":
            used_extensions = self.build_extension_list(event)

        if len(filename_hint) == 6:
            tempfile, tempfile_len = self.generate_templist(filename_hint, shortname_type)
            self.verbose(
                f"generated temp word list of size [{str(tempfile_len)}] for filename hint: [{filename_hint}]"
            )

        else:
            tempfile = self.helpers.tempfile([filename_hint], pipe=False)
            tempfile_len = 1

        if tempfile_len > 0:
            if shortname_type == "endpoint":
                for ext in used_extensions:
                    async for r in self.execute_ffuf(tempfile, root_url, suffix=f".{ext}"):
                        await self.emit_event(
                            r["url"],
                            "URL_UNVERIFIED",
                            parent=event,
                            tags=[f"status-{r['status']}"],
                            context=f"{{module}} brute-forced {ext.upper()} files at {root_url} and found {{event.type}}: {{event.data}}",
                        )

            elif shortname_type == "directory":
                async for r in self.execute_ffuf(tempfile, root_url, exts=["/"]):
                    r_url = f"{r['url'].rstrip('/')}/"
                    await self.emit_event(
                        r_url,
                        "URL_UNVERIFIED",
                        parent=event,
                        tags=[f"status-{r['status']}"],
                        context=f"{{module}} brute-forced directories at {r_url} and found {{event.type}}: {{event.data}}",
                    )

        if self.config.get("find_delimiters"):
            if "shortname-directory" in event.tags:
                delimiter_r = self.find_delimiter(filename_hint)
                if delimiter_r:
                    delimiter, prefix, partial_hint = delimiter_r
                    self.verbose(f"Detected delimiter [{delimiter}] in hint [{filename_hint}]")
                    tempfile, tempfile_len = self.generate_templist(partial_hint, "directory")
                    ffuf_prefix = f"{prefix}{delimiter}"
                    async for r in self.execute_ffuf(tempfile, root_url, prefix=ffuf_prefix, exts=["/"]):
                        await self.emit_event(
                            r["url"],
                            "URL_UNVERIFIED",
                            parent=event,
                            tags=[f"status-{r['status']}"],
                            context=f'{{module}} brute-forced directories with detected prefix "{ffuf_prefix}" and found {{event.type}}: {{event.data}}',
                        )

            elif "shortname-endpoint" in event.tags:
                for ext in used_extensions:
                    delimiter_r = self.find_delimiter(filename_hint)
                    if delimiter_r:
                        delimiter, prefix, partial_hint = delimiter_r
                        self.verbose(f"Detected delimiter [{delimiter}] in hint [{filename_hint}]")
                        tempfile, tempfile_len = self.generate_templist(partial_hint, "endpoint")
                        ffuf_prefix = f"{prefix}{delimiter}"
                        async for r in self.execute_ffuf(tempfile, root_url, prefix=ffuf_prefix, suffix=f".{ext}"):
                            await self.emit_event(
                                r["url"],
                                "URL_UNVERIFIED",
                                parent=event,
                                tags=[f"status-{r['status']}"],
                                context=f'{{module}} brute-forced {ext.upper()} files with detected prefix "{ffuf_prefix}" and found {{event.type}}: {{event.data}}',
                            )

        if self.config.get("find_subwords"):
            subword, suffix = self.find_subword(filename_hint)
            if subword:
                if "shortname-directory" in event.tags:
                    tempfile, tempfile_len = self.generate_templist(suffix, "directory")
                    async for r in self.execute_ffuf(tempfile, root_url, prefix=subword, exts=["/"]):
                        await self.emit_event(
                            r["url"],
                            "URL_UNVERIFIED",
                            parent=event,
                            tags=[f"status-{r['status']}"],
                            context=f'{{module}} brute-forced directories with detected subword "{subword}" and found {{event.type}}: {{event.data}}',
                        )
                elif "shortname-endpoint" in event.tags:
                    for ext in used_extensions:
                        tempfile, tempfile_len = self.generate_templist(suffix, "endpoint")
                        async for r in self.execute_ffuf(tempfile, root_url, prefix=subword, suffix=f".{ext}"):
                            await self.emit_event(
                                r["url"],
                                "URL_UNVERIFIED",
                                parent=event,
                                tags=[f"status-{r['status']}"],
                                context=f'{{module}} brute-forced {ext.upper()} files with detected subword "{subword}" and found {{event.type}}: {{event.data}}',
                            )

    async def finish(self):
        if self.config.get("find_common_prefixes"):
            per_host_collection = dict(self.per_host_collection)
            self.per_host_collection.clear()

            for host, hint_tuple_list in per_host_collection.items():
                hint_list = [x[0] for x in hint_tuple_list]

                common_prefixes = self.find_common_prefixes(hint_list)
                for prefix in common_prefixes:
                    self.verbose(f"Found common prefix: [{prefix}] for host [{host}]")
                    for hint_tuple in hint_tuple_list:
                        hint, url = hint_tuple
                        if hint.startswith(prefix):
                            if "shortname-endpoint" in self.shortname_to_event[hint].tags:
                                shortname_type = "endpoint"
                            elif "shortname-directory" in self.shortname_to_event[hint].tags:
                                shortname_type = "directory"
                            else:
                                self.error("ffuf_shortnames received URL_HINT without proper 'shortname-' tag")
                                continue

                            partial_hint = hint[len(prefix) :]

                            # safeguard to prevent loading the entire wordlist
                            if len(partial_hint) > 0:
                                tempfile, tempfile_len = self.generate_templist(partial_hint, shortname_type)

                                if "shortname-directory" in self.shortname_to_event[hint].tags:
                                    self.verbose(
                                        f"Running common prefix check for URL_HINT: {hint} with prefix: {prefix} and partial_hint: {partial_hint}"
                                    )

                                    async for r in self.execute_ffuf(tempfile, url, prefix=prefix, exts=["/"]):
                                        await self.emit_event(
                                            r["url"],
                                            "URL_UNVERIFIED",
                                            parent=self.shortname_to_event[hint],
                                            tags=[f"status-{r['status']}"],
                                            context=f'{{module}} brute-forced directories with common prefix "{prefix}" and found {{event.type}}: {{event.data}}',
                                        )
                                elif shortname_type == "endpoint":
                                    used_extensions = self.build_extension_list(self.shortname_to_event[hint])

                                    for ext in used_extensions:
                                        self.verbose(
                                            f"Running common prefix check for URL_HINT: {hint} with prefix: {prefix}, extension: .{ext}, and partial_hint: {partial_hint}"
                                        )
                                        async for r in self.execute_ffuf(
                                            tempfile, url, prefix=prefix, suffix=f".{ext}"
                                        ):
                                            await self.emit_event(
                                                r["url"],
                                                "URL_UNVERIFIED",
                                                parent=self.shortname_to_event[hint],
                                                tags=[f"status-{r['status']}"],
                                                context=f'{{module}} brute-forced {ext.upper()} files with common prefix "{prefix}" and found {{event.type}}: {{event.data}}',
                                            )



import json
import httpx

from .base import ModuleTestBase


class TestSplunk(ModuleTestBase):
    downstream_url = "https://splunk.blacklanternsecurity.fakedomain:1234/services/collector"
    config_overrides = {
        "modules": {
            "splunk": {
                "url": downstream_url,
                "hectoken": "HECTOKEN",
                "index": "bbot_index",
                "source": "bbot_source",
            }
        }
    }

    def verify_data(self, j):
        if not j["source"] == "bbot_source":
            return False
        if not j["index"] == "bbot_index":
            return False
        data = j["event"]
        if not data["data"] == "blacklanternsecurity.com" and data["type"] == "DNS_NAME":
            return False
        return True

    async def setup_after_prep(self, module_test):
        self.url_correct = False
        self.method_correct = False
        self.got_event = False
        self.headers_correct = False

        async def custom_callback(request):
            j = json.loads(request.content)
            if request.url == self.downstream_url:
                self.url_correct = True
            if request.method == "POST":
                self.method_correct = True
            if "Authorization" in request.headers:
                self.headers_correct = True
            if self.verify_data(j):
                self.got_event = True
            return httpx.Response(
                status_code=200,
            )

        module_test.httpx_mock.add_callback(custom_callback)
        module_test.httpx_mock.add_callback(custom_callback)
        module_test.httpx_mock.add_response()

    def check(self, module_test, events):
        assert self.got_event is True
        assert self.headers_correct is True
        assert self.method_correct is True
        assert self.url_correct is True
